!+
! Module: epatplot_mod
!
! Authors:  Michael Freyberg - MPE Garching
!           Frank Haberl     - MPE Garching
!           Konrad Dennerl   - MPE Garching
!          
! Description: Plot EPIC pn and EPIC MOS event pattern statistics
!
! Comments:
!
!-

module epatplot_mod

  use eventAttributes
  use cal
  use caloaldefs
  use dal
  use param
  use errorHandling

  implicit none

  private

  public epatplot

contains

!---------------------------------------------------------------------
subroutine epatplot
!---------------------------------------------------------------------

  implicit none

  character(len=10)  ,parameter  :: CodeVersion = '1.1.6'
  integer(kind=Int32),parameter  :: maxbin = 4096, maxbin_times_2 = 2*maxbin
  integer(kind=Int32),parameter  :: maxpat = 4, maxpat_plus_1 = maxpat+1, &
                                    maxpat_plus_2 = maxpat+2
  integer(kind=Int32),parameter  :: maxmodel = maxpat
! integer                        :: statusPG, pgopen
!
  logical(kind=bool)  :: enable_qdp = .false., &
                         enable_pum = .false., &
                         enable_y   = .false.
!
  real(kind=single)   :: h, he, hs, hd, ht, hq, sigma, s,              &
                         xmin, xmax, ymin, ymax, xtxt, ytxt, yt0, yt1, &
                         xdat, ydat, xsiz, ysiz,                       &
                         xamin, xamax, xasiz,                          &
                         vxmin, vxmax, vymin, vymax, grey,             &
                         sprb(maxbin), rhis(maxpat,maxbin),            &
                         xhis(1:maxpat_plus_1,1:maxbin_times_2),       &
                         yhis(1:maxpat_plus_1,1:maxbin_times_2),       &
                         model(maxmodel,maxbin), livetime, y, energy(maxbin)
  integer(kind=Int32) :: ntrow, nsrow, i, j, k, npos, mpos, fpos,      &
                         nhas, mhas, om1, om2,                         &
                         ix0, ix, iy0, iy1, iy, iex, irow,             &
                         mbin, kbin, ibin, ks, kd, kt, kq, kpat,       &
                         khis(1:maxpat_plus_2,1:maxbin),               &
                         jm(maxbin), jp(maxbin),                       &
                         ipat, jpat, jcts(maxbin),                     &
                         icol(maxpat_plus_1), k1, k2, nbin(maxpat_plus_1)
  integer(kind=Int32) :: CALstat_mode, CALstat_node, CALstat_ccd, &
                         CALstat_inst
  integer(kind=Int16) :: ix16, iy16
  integer(kind=Int32), dimension(2)  :: imageDim
  integer             :: iw, ierr, dthresh, mthresh(0:5), mmp,    &
                         calMode, userMode, node = 0, revolut,    &
                         nccd = 12, iccd, quadrant, ccdid,        &
                         AccessType = READ, EventsFileType,       &
                         xw1 = 1, xw2 = 1, nx, ndx,               &
                         yw1 = 1, yw2 = 1, ny, ndy
  logical(kind=bool)  :: is_ps = .false., is_gif = .false.,       &
                         flag0 = .true., hasCCD(1:12,0:1),        &
                         is_epn = .false., is_mos = .false., is_rgs = .false.
  logical(kind=bool)  :: xaxisadu = .false., useplotfile = .false.,   &
                         withflag = .true., withoutputmask = .false., &
                         badpix_from_events = .true.,                 &
                         offsets_from_events = .true.,                &
                         hasBADPIX(0:12,0:1) = .false.,               &
                         hasOFFSET(0:12,0:1) = .false.,               &
                         withccdnr = .true., singleChip = .false. ,   &
                         lunrecogn = .false.
  character(len=1)    :: chpt(maxpat_plus_1)
  character(len=4)    :: ccdmode, mmode(0:5)
  character(len=8)    :: blockname(0:12),blockname2(0:12), revlab
  character(len=40)   :: instmode, device, ext, xlabel, instrument,   &
                         insttxt, rawytxt, filter
  character(len=1024) :: infilename, outfilename, canonicalname,      &
                         outdir, plotfile, outqdpname, outmaskname,   &
                         omsk1, omsk2
!
  integer(kind=int16),dimension(:,:), pointer     :: FOVmap
  integer(kind=Int8), dimension(:,:), pointer     :: map_arrayP
  integer(kind=Int8), dimension(:,:), allocatable :: map_array

  type(DataSetT)      :: infile, outmask
  type(ArrayT)        :: map
  type(TableT)        :: events, badpix, offset
  type(ColumnT)       :: pattern,pi,flag,rawx,rawy,ccdnr,yextent
  type(PatternFractionServerT)    :: pfds
  type(CurveT)                    :: curve
  integer(kind=Int8),  dimension(:), pointer :: patternCol,ccdnrCol
  integer(kind=Int16), dimension(:), pointer :: piCol,rawxCol,rawyCol, &
                                                yextentCol
  integer(kind=Int32), dimension(:), pointer :: flagCol
  real(double), dimension(:), pointer        :: pattFracChan
!
  real(kind=single)           ::  conv = 1.0, userY = 190.0
  real(kind=single)           ::  xpos = 32., ypos, srcpos
! real(kind=single),parameter ::  conv = 5.e-3  ! keV/adu
! real(kind=single),parameter ::  conv = 1.0    ! adu channel
  real(kind=single),parameter ::  econv = 5.0   ! eV per adu channel
!
  integer(kind=int8),parameter :: nband = 6
  real(kind=single)           ::  e_low(1:nband),e_hig(1:nband), &
                                  e_sum(1:nband,1:maxpat_plus_2)

  data e_low    / 200,0500,2000,4500,07500,01000 /
  data e_hig    / 500,2000,4500,7500,12000,32800 /
  data icol     / 2, 4, 3, 5, 18 /
  data mthresh  / 40,40,40,40,80,46 /
  data mMode    / 'FF','eFF','LW','SW','TI','BU' /
  data chpt     / 's', 'd', 't', 'q', 'u' /
!
!-----------------------------------------------------------------------
! start the program:
! -------------------------------------------
!
  write(errstr,*) 'epatplot ',trim(CodeVersion),' is running...'
  call message(AppMsg,Sparse)

! -------------------------------------------
! initialize the histograms:
! -------------------------------------------

  khis(1:maxpat_plus_2,1:maxbin) = 0
  xhis(1:maxpat_plus_1,1:maxbin_times_2) = 0
  yhis(1:maxpat_plus_1,1:maxbin_times_2) = 0
  rhis(1:maxpat,1:maxbin) = 0.0
  energy(1:maxbin) = 0.0
  model(1:maxmodel,1:maxbin) = 0.0
  jm(1:maxbin) = 0
  jp(1:maxbin) = 0
  jcts(1:maxbin) = 0

! -------------------------------------------
! get input parameters:
! -------------------------------------------

  call getInParamValues(infilename, sigma, device, outdir, withflag, &
                        useplotfile, plotfile, xaxisadu, enable_qdp, &
                        enable_pum, userMode, enable_y, userY, &
                        withoutputmask,outmaskname)

  call stripSlash(infilename)
  call getCanonicalName(infilename, canonicalname)
  call stripSlash(canonicalname)
  call stripSlash(plotfile)

  if (withoutputmask) then
    om1 = index(outmaskname,'##')
    om2 = om1+2
    write(errstr,'(1x,a,t35,i9)') 'CCD placeholder ## at position: ',om1
    call message(AppMsg,Noisy)
    if (om1>1) then
      omsk1 = outmaskname(1:om1-1)
      omsk2 = outmaskname(om2:)
    elseif (om1 == 1) then
      omsk1 = ""
      omsk2 = outmaskname(om2:)
    else
      omsk1 = outmaskname
      omsk2 = ""
    endif
  endif

  select case (device)
    case ('/VCPS')
      ext = 'ps'
    case ('/GIF')
      ext = 'gif'
    case default
      ext = 'ps'
  end select

  if (useplotfile) then
    outfilename = trim(outdir) // '/' // trim(plotfile)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  else
    outfilename = trim(outdir) // '/' // trim(canonicalname) // '.' // trim(ext)
    outqdpname  = trim(outdir) // '/' // trim(canonicalname) // '.qdp'
  endif

  call stripSlash(outfilename)
  call stripSlash(outqdpname)

! print*,'|',trim(outfilename),'|'

  write(errstr,'(1x,a,a)') 'Accessing input data set: ',trim(infilename)
  call message(AppMsg,Verbose)

  if (.not.setexists(infilename)) then
    call error("noSet","Input data set not found")
  endif
  infile = dataSet(infilename,READ,LOW_MEMORY)

  if (.not. hasBlock(infile,'EVENTS')) then    ! no EVENTS extension
    call error("noEvents","No EVENTS extension in input data set")
  endif
  events = table(infile,'EVENTS',READ)

!
! read FILTER attribute and fill relevant text string
!
  if (hasAttribute(infile,"FILTER")) then
    filter = trim(stringAttribute(infile,"FILTER"))
    write(errstr,'(1x,a,t39,a)') 'Filter:     ',trim(filter)
    call message(AppMsg,Verbose)
  else
    filter = ' '
    call message(AppMsg,Verbose,'FILTER attribute not found')
  endif

!
! read INSTRUME attribute and fill relevant text string
!
  if (hasAttribute(events,"INSTRUME")) then
    instrument = trim(stringAttribute(events,"INSTRUME"))
    write(errstr,'(1x,a,t39,a)') 'Instrument: ',trim(instrument)
    call message(AppMsg,Verbose)
  else
    instrument = ' '
    call message(AppMsg,Verbose,'INSTRUME attribute not found')
  endif

  select case (instrument)
    case ('EPN')
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
    case ('EMOS1')
      insttxt = 'EPIC MOS1'
      is_mos = .true.
      CALstat_inst = EMOS1
    case ('EMOS2')
      insttxt = 'EPIC MOS2'
      is_mos = .true.
      CALstat_inst = EMOS2
    case ('RGS1')
      insttxt = 'RGS1'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS1
    case ('RGS2')
      insttxt = 'RGS2'
      nccd = 9
      is_rgs = .true.
      CALstat_inst = RGS2
    case default
      insttxt = 'EPIC pn'
      is_epn = .true.
      CALstat_inst = EPN
  end select

  if (.not.is_epn) then
    if (xaxisadu) then
      call warning("isnotEPN","xaxisadu=Y only implemented for EPIC pn.")
      xaxisadu = .false.
    endif
  endif
  if (xaxisadu) then
    conv = 1.0
    if (is_epn) then
      xlabel = 'ADU Channel [5 eV]'
    else
      xlabel = 'ADU Channel'
    endif
  else
    conv = 5.0    ! channel -> eV
    xlabel = 'PI Channel [eV]'
  endif

!
! get PI energy of ADU channel
!
  if (is_epn) then
    do i = 1,maxbin
      energy(i) = real(i)*5.00 ! channel -> eV
    enddo
    nccd = 12
    ndx = EPN_MAX_X
    ndy = EPN_MAX_Y
    yt0 = 0.1
    yt1 = 0.2
  elseif (is_mos) then
    do i = 1,maxbin
      energy(i) = real(i)*3.25 ! channel -> pseudo-energy
    enddo
    nccd = 7
    ndx = EMOS_MAX_X
    ndy = EMOS_MAX_Y
    mMode(0:5) = 'MOS1'
    yt0 = 0.125
    yt1 = 0.25
  else
    call warning("isnotEPIC","only EPIC model curves are implemented yet")
    withoutputmask = .false.   ! only implemented for EPIC at the moment
    do i = 1,maxbin
      energy(i) = real(i)*1.00 ! channel -> pseudo-energy
    enddo
    ndx = 1
    ndy = 1
    yt0 = 0.125
    yt1 = 0.25
  endif
  imageDim(1) = ndx
  imageDim(2) = ndy
  nx =  ndx
  ny =  ndy
!
  if (withoutputmask) allocate(map_array(1:ndx,1:ndy))
!
! access events file columns
!
  ntrow = numberOfRows(events)
  if (ntrow .le. 0) then
    call error('emptyInput','No events in input file.')
  else
    write(errstr,'(1x,a,t35,i9)') 'Number of events in input file: ',ntrow
    call message(AppMsg,Verbose)    
  endif

  if (.not.hasColumn(events,'PATTERN')) then
    call error("noPattern","No PATTERN column in input data set")
  endif
  pattern = column(events,'PATTERN',READ)
  patternCol => int8data(pattern)

  if (.not.hasColumn(events,'PI')) then
    call error("noPI","No PI column in input data set")
  endif
  pi = column(events,'PI',READ)
  piCol => int16data(pi)

  if (.not.hasColumn(events,'RAWY')) then
    call error("noRAWY","No RAWY column in input data set")
  endif
  rawy = column(events,'RAWY',READ)
  rawyCol => int16data(rawy)

  if (.not.hasColumn(events,'RAWX')) then
    call error("noRAWX","No RAWX column in input data set")
  endif
  rawx = column(events,'RAWX',READ)
  rawxCol => int16data(rawx)

  if (.not.hasColumn(events,'FLAG')) then
    if (withflag) then
      call warning("noFlag", &
          "No FLAG column in input data set, using all events for plots.")
      flag0 = .true.
      withflag = .false.
    endif
  else
    flag = column(events,'FLAG',READ)
    flagCol => int32data(flag)
  endif

  if (.not.hasColumn(events,'CCDNR')) then
    if (withccdnr) then
      call warning("noCCDnr", &
          "No CCDNR column in input data set, will check all CCDs.")
      withccdnr = .false.
    endif
  else
    ccdnr = column(events,'CCDNR',READ)
    ccdnrCol => int8data(ccdnr)
  endif

  if (hasAttribute(events,"SUBMODE")) then
    instmode = trim(stringAttribute(events,"SUBMODE"))
  else
    instmode = ' '
    call message(AppMsg,Verbose,'SUBMODE attribute not found')
  endif

  if (hasAttribute(events,"LIVETIME")) then
    livetime = real64Attribute(events,"LIVETIME")
    write(errstr,'(1x,a,t35,f9.1,a)') 'Livetime by evselect: ',livetime,' [s]'
    call message(AppMsg,Verbose)
  else
    livetime = 0.0
  endif

  if (hasAttribute(events,"REVOLUT")) then
    revolut = int16Attribute(events,"REVOLUT")
    write(revlab,'(a,i4.4)') 'Rev.',revolut
  else
    revlab = ' '
  endif

  select case (instmode)
    case ('PrimeFullWindow')
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
    case ('PrimeFullWindowExtended')
      calMode = 1
      CALstat_Mode = PRIME_FULL_WINDOW_EXTENDED
    case ('PrimeLargeWindow')
      calMode = 2
      CALstat_Mode = PRIME_LARGE_WINDOW
    case ('PrimeSmallWindow')
      calMode = 3
      CALstat_Mode = PRIME_SMALL_WINDOW
    case ('FastTiming')
      calMode = 4
      CALstat_Mode = FAST_TIMING
    case ('FastBurst')
      calMode = 5
      CALstat_Mode = FAST_BURST
    case default
      ccdmode = '  '
      calMode = 0
      CALstat_Mode = PRIME_FULL_WINDOW
      dthresh = 40
  end select

  if (enable_pum) calMode = userMode
  dthresh = mthresh(calMode)
  ccdmode = mmode(calMode)
  write(errstr,'(1x,a,t35,a)') 'Instrument mode of observation:',trim(instmode)
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,f9.1)') 'Sigma for rebinning: ',sigma
  call message(AppMsg,Verbose)


! +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  call message(AppMsg,Noisy,'Reading input data set ...')

  npos = 0
  mpos = 0
  fpos = 0
  do i = 1, ntrow
    if (withccdnr) then
      iccd = ccdnrCol(i-1)
      if ((iccd.ge.1).and.(iccd.le.nccd)) then
        hasCCD(iccd,node) = .true.
      endif
    endif
    jpat = patternCol(i-1)
    if (withflag) flag0 = (flagCol(i-1) == 0)
    if (is_EPN) then
      lunrecogn = ((((iand(flagCol(i-1),EVATT_INVALID_PATTERN)>0).and.(jpat>4)) .or. &
                  (jpat>12)) .and. piCol(i-1)>0 .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_CCD_WINDOW)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_ONBOARD_BADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_BRIGHTPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_CLOSE_TO_DEADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_COSMIC_RAY)==0) .and. &
                  (iand(flagCol(i-1),EVATT_TRAILING)==0) .and. &
                  (iand(flagCol(i-1),EVATT_ON_BADPIX)==0) .and. &
                  (iand(flagCol(i-1),EVATT_SECONDARY)==0) .and. &
                  (iand(flagCol(i-1),EVATT_MIP_ASSOCIATED)==0) .and. &
                  (iand(flagCol(i-1),EVATT_OUT_OF_CCD_WINDOW)==0) .and. &
                  (iand(flagCol(i-1),EVATT_OUT_OF_FOV)==0) .and. &
                  (iand(flagCol(i-1),EVATT_IN_BAD_FRAME)==0) .and. &
                  (iand(flagCol(i-1),EVATT_IN_SPOILED_FRAME)==0))
    elseif (is_MOS) then
      lunrecogn = (jpat.eq.31 .and. flag0)
    else
      lunrecogn = .FALSE.
    endif
    if (jpat.eq.0 .and. flag0) then                       ! singles
      kpat = 1
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.1.and.jpat.le.4 .and. flag0) then     ! doubles
      kpat = 2
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.5.and.jpat.le.8 .and. flag0) then     ! triples
      kpat = 3
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
    elseif (jpat.ge.9.and.jpat.le.12 .and. flag0) then    ! quadruples
      kpat = 4
      ypos = ypos + rawyCol(i-1)
      npos = npos+1
!   elseif (jpat.le.12 .and. .not.flag0) then    ! not FLAG=0
!   elseif (rawyCol(i-1).eq.200) then    !
!     kpat = 6
!     fpos = fpos+1      
!   else
!   elseif ((iand(flagCol(i-1),EVATT_INVALID_PATTERN)>0).or.(jpat>12)) then
    elseif (lunrecogn) then
      kpat = 5
      mpos = mpos+1
      if (verbosity(AppMsg) >= Noisy) then
        if (piCol(i-1)<220) then
          write(errstr,'(i9,i5,i12,i6,i3,i4)') &
                i,jpat,flagCol(i-1),piCol(i-1), &
                rawxcol(i-1),rawycol(i-1)
          call message(AppMsg,Sparse)
        endif
      endif
    else
      kpat=6
    endif
!
! accumulate energies of pattern types in bands
!
    if (kpat .le. maxpat_plus_1) then
      do j = 1,nband
        if (piCol(i-1).ge.e_low(j).and.piCol(i-1).lt.e_hig(j)) then
          e_sum(j,kpat) = e_sum(j,kpat) + piCol(i-1)
        endif
      enddo
    endif
!
! find the energy bin corresposnding to event PI channel
!
    kbin = piCol(i-1)/econv                               ! PN only !!

    if (kpat.ge.1 .and. kpat.le.maxpat_plus_1 .and.   &
        kbin.ge.1 .and. kbin.le.maxbin) then
      khis(kpat,kbin) = khis(kpat,kbin) + 1
    endif
  enddo

  write(errstr,'(1x,a,t35,12l1)') 'CCDs in events file: ', &
                (hasCCD(iccd,node),iccd=1,nccd)
  call message(AppMsg,Verbose)

  write(errstr,'(1x,a,t35,i9  )') 'Number of recognized photons: ',npos 
  call message(AppMsg,Verbose)
  write(errstr,'(1x,a,t35,i9  )') 'Number of unrecognized cases: ',mpos 
  call message(AppMsg,Verbose)


  do j = 1,nband
    do kpat = 1,maxpat_plus_1
      e_sum(j,maxpat_plus_2) = e_sum(j,maxpat_plus_2)+e_sum(j,kpat)
    enddo
  enddo

  do j = 1,nband
    write(errstr,'(1x,i1,2f8.1,5e11.4)') j,e_low(j),e_hig(j),&
         (e_sum(j,kpat),kpat=1,maxpat_plus_1)
!   call message(AppMsg,Sparse)
  enddo
  call message(AppMsg,Verbose,' band   E_low   E_hig frac_inv') 
  do j = 1,nband
    write(errstr,'(4x,i1,2f8.1,f8.5)') j,e_low(j),e_hig(j),&
         e_sum(j,maxpat_plus_1)/e_sum(j,maxpat_plus_2)
    call message(AppMsg,Verbose)       
  enddo


  if (npos > 0) then
    ypos = ypos/npos
  else
    ypos = 190
  endif

  if ((calMode == 4 .or. calMode == 5).and.(is_EPN)) then
    if (hasAttribute(events,"SRCPOS")) then
      srcpos = int16Attribute(events,"SRCPOS")
    else
      srcpos = 190
      call message(AppMsg,Sparse,'SRCPOS attribute not found')
    endif
    ypos = srcpos
  endif

  write(errstr,'(1x,a,t35,f9.3)') 'Average RAWY position: ',ypos 
  call message(AppMsg,Verbose)

  if (enable_Y) then
    write(errstr,'(1x,a,t35,f9.3)') 'User-defined RAWY position: ',userY 
    call message(AppMsg,Verbose)
    ypos = userY
  endif

  if (ypos < 10) then
    write(rawytxt,'(a,f3.1)') 'RAWY = ',ypos
  elseif (ypos < 100) then
    write(rawytxt,'(a,f4.1)') 'RAWY = ',ypos
  else
    write(rawytxt,'(a,f5.1)') 'RAWY = ',ypos
  endif

  if (withflag) call release(flag)
  if (withccdnr) call release(ccdnr)
  call release(pattern)
  call release(pi)
  call release(rawx)
  call release(rawy)

!
! output mask stuff (FLAG==0)
!
  if (withoutputmask) then

    node = 0
    nhas = 0
    mhas = 0
    CALstat_node = node
    EventsFileType = 0

    hasBADPIX(0:nccd,0:1) = .false.
    hasOFFSET(0:nccd,0:1) = .false.

    do iccd = 0,nccd
      if (iccd == 0) then
        write(blockname(iccd), '(a8)') 'BADPIX  '
        write(blockname2(iccd),'(a8)') 'OFFSETCO'
      else
        write(blockname(iccd), '(a6,i2.2)') 'BADPIX',iccd
        write(blockname2(iccd),'(a6,i2.2)') 'OFFSET',iccd
      endif

      if (hasBlock(infile,blockname(iccd))) then
        hasBADPIX(iccd,node) = .true.
        if (iccd>0) nhas = nhas+1
      endif
      if (hasBlock(infile,blockname2(iccd))) then
        hasOFFSET(iccd,node) = .true.
        if (iccd>0) mhas = mhas+1
      endif
    enddo

    if (hasBADPIX(0,0)) then
      singleChip = .true.
      nhas = 1
    elseif (nhas > 0) then
      singleChip = .false.
    else
      singleChip = .false.
      badpix_from_events = .false.
    endif

    if (singleChip) then
      if (hasBlock(infile,'OFFSETCO')) then      ! old-style single chip file
        EventsFileType = 1
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style single chip file
        EventsFileType = 2
        blockname2(0) = 'OFFSETS '               ! -> redefine extension name
        hasOFFSET(0,node) = .true.
        mhas = 1
      else                                       ! very old single chip file
        EventsFileType = 6                       ! -> not supported
        offsets_from_events = .false.
      endif
    else
      if (mhas > 0) then                         ! old-style epchain event file
        EventsFileType = 3
      elseif (hasBlock(infile,'OFFSETS')) then   ! new-style epchain event file
        EventsFileType = 4
        blockname2(1:nccd) = 'OFFSETS '          ! -> redefine extension names
        hasOFFSET(1:nccd,node) = .true.
        mhas = 1
      else                                       ! old-style LUX/PPS event file
        EventsFileType = 5                       ! -> not supported
        offsets_from_events = .false.
      endif
    endif
    write(errstr,'(1x,a,t35,i9,2x,a,l1,a)') 'Event file type (single): ', &
         EventsFileType,' (',singleChip,')'
    call message(AppMsg,Noisy)
    write(errstr,'(1x,a,t35,i9,i5)') 'Extensions for BADPIX and OFFSET: ', &
         nhas,mhas
    call message(AppMsg,Noisy)

    if (nhas == 0) then    ! no BADPIX extension
      call warning("noBadpix","No BADPIX extension in input data set")
      badpix_from_events = .false.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from CCF instead')
      call message(AppMsg,Verbose,' This is not yet implemented ! No action performed.')
    else
      badpix_from_events = .true.
      call message(AppMsg,Verbose,' Output mask will use BADPIX from events file')
    endif

    if (badpix_from_events) then
      do iccd = 0, nccd
        map_array(1:ndx,1:ndy) = 1
        CALstat_ccd = iccd
        nsrow = 0
        if (hasBADPIX(iccd,node)) then
          badpix = table(infile,blockname(iccd),AccessType)
          if (hasAttribute(badpix,"WINDOWX0")) then
            xw1 = int16Attribute(badpix,"WINDOWX0")
          else
            call message(AppMsg,Verbose,'WINDOWX0 attribute not found')
          endif
          if (hasAttribute(badpix,"WINDOWDX")) then
            nx = int16Attribute(badpix,"WINDOWDX")
          else
            call message(AppMsg,Verbose,'WINDOWDX attribute not found')
          endif
          xw2 = xw1+nx-1

          if (hasAttribute(badpix,"WINDOWY0")) then
            yw1 = int16Attribute(badpix,"WINDOWY0")
          else
            call message(AppMsg,Verbose,'WINDOWY0 attribute not found')
          endif
          if (hasAttribute(badpix,"WINDOWDY")) then
            ny = int16Attribute(badpix,"WINDOWDY")
          else
            call message(AppMsg,Verbose,'WINDOWDY attribute not found')
          endif
          yw2 = yw1+ny-1
          write(errstr,'(1x,a,t35,i9,5(2x,i5))') 'Window parameters: ', &
                xw1,nx,xw2,yw1,ny,yw2
          call message(AppMsg,Noisy)

          nsrow = numberOfRows(badpix)
          write(errstr,'(1x,a,t35,i9)') 'Number of bad pixel entries: ', nsrow
          call message(AppMsg,Noisy)
          if (nsrow > 0) then
            rawx = column(badpix,"RAWX",AccessType)
            rawy = column(badpix,"RAWY",AccessType)
            yextent = column(badpix,"YEXTENT",AccessType)
            rawxCol => int16Data(rawx)
            rawyCol => int16Data(rawy)
            yextentCol => Int16Data(yextent)
          endif
          if (singleChip .and. iccd == 0) then
            if (is_epn) then
              quadrant = int8attribute(badpix,"QUADRANT")
              ccdid = int8attribute(badpix,"CCDID")
              CALstat_ccd = 3*quadrant+ccdid+1            
              call CAL_setState(CALstat_inst,              &
                    CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node, 
                    accuracyLevel = ACCURACY_HIGH)
            elseif (is_mos) then
              ccdid = int8attribute(badpix,"CCDID")
              CALstat_ccd = ccdid            
              call CAL_setState(CALstat_inst,              &
                    CALstat_ccd, CALstat_node, ccdModeId = CALstat_mode, &
                    accuracyLevel = ACCURACY_HIGH)
            endif
          elseif (.not.singleChip .and. iccd > 0) then
            call CAL_setState(CALstat_inst,    &
                 CALstat_ccd, ccdModeId = CALstat_mode, &   ! CALstat_node, 
                 accuracyLevel = ACCURACY_HIGH)
          else
            call message(AppMsg,Sparse,' ... BADPIX and BADPIXnn extension ...')
          endif

          call message(AppMsg,Noisy,' ... FLAG mask: BADPIX ...')
          do irow = 0,nsrow-1
            iex = yextentCol(irow)
            ix0 = rawxCol(irow)
            iy0 = rawyCol(irow)
            iy1 = min(iy0+iex-1,ndy)
            map_array(ix0,iy0) = 0
            do iy = iy0,iy1
              map_array(ix0,iy) = 0
              if (ix0.gt.1) map_array(ix0-1,iy) = 0
              if (ix0.lt.ndx) map_array(ix0+1,iy) = 0
            enddo
            if (iy0.gt.1) map_array(ix0,iy0-1) = 0
            if (iy1.lt.ndy) map_array(ix0,iy1+1) = 0
          enddo

          if (nsrow > 0) then
            call release(rawx)
            call release(rawy)
            call release(yextent)
          endif
!
! set FOV map
!
          if ((calMode == 4 .or. calMode == 5).and.(is_EPN)) then
            call message(AppMsg,Noisy,' ... FLAG mask: FOVmap ... not done for TI+BU mode')
          else
            call message(AppMsg,Noisy,' ... FLAG mask: FOVmap ...')
            call CAL_getFOVmap(FOVmap)
            do iy0 = 1,ndy
              iy = iy0-1
              do ix0 = 1,ndx
                ix = ix0-1
                map_array(ix0,iy0) = map_array(ix0,iy0)*FOVmap(ix,iy)
              enddo
            enddo
          endif
     !
     ! set read-out window map
     !
          call message(AppMsg,Noisy,' ... FLAG mask: WINDOW ...')
          if (xw1 .ge. 1 .and. xw2 .ge. xw1 .and. &
            yw1 .ge. 1 .and. yw2 .ge. yw1 ) then
              map_array(1:xw1,  1:ndy)   = 0
            map_array(xw2:ndx,1:ndy)   = 0
            map_array(1:ndx,  1:yw1)   = 0
            map_array(1:ndx,  yw2:ndy) = 0
          else
             write(errstr,*) 'Window parameters: XW1,NX,XW2,YW1,NY,YW2 = ', &
                              xw1,nx,xw2,yw1,ny,yw2
            call warning('invalidWindow')
          endif
     !
     ! set offset column map
     !
          call message(AppMsg,Noisy,' ... FLAG mask: OFFSET ...')
          if (hasOFFSET(iccd,node)) then
            offset = table(infile,blockname2(iccd),AccessType)
            nsrow = numberOfRows(offset)
            write(errstr,'(1x,a,t35,i9)') 'Number of offset columns: ', nsrow
            call message(AppMsg,Noisy)
            if (nsrow > 0) then
              rawx = column(offset,"RAWX",AccessType)
              rawxCol => int16Data(rawx)
              ccdnr = column(offset,'CCDNR',READ)
              ccdnrCol => int8data(ccdnr)

              do irow = 0,nsrow-1
                if (ccdnrCol(irow) == CALstat_ccd) then
                  ix = rawxCol(irow)
                  map_array(ix,1:ndy) = 0
                  if (ix.gt.1)   map_array(ix-1,1:ndy) = 0
                  if (ix.lt.ndx) map_array(ix+1,1:ndy) = 0
                endif
              enddo

              call release(rawx)     ! column
              call release(ccdnr)    ! column
            endif
            call release(offset)   ! extension
          else
            write(errstr,*) 'Does not have OFFSET for CCD = ',iccd,' [',&
                      trim(blockname2(iccd)),']'
            call message(AppMsg,Noisy)   
          endif
 
          write(outmaskname,'(a,i2.2,a)') trim(omsk1),CALstat_ccd,trim(omsk2)
          write(errstr,*) '... FLAG mask: ',trim(outmaskname)
          call message(AppMsg,Verbose)
          outmask = dataSet(outmaskname,Create)
          map = addArray(outmask, "PRIMARY", Integer8, imageDim)
          map_arrayP => int8Array2Data(map)
          map_arrayP = map_array
          call setAttributes(attributable(map),attributable(badpix))
          if (hasAttribute(map,"HISTORY")) then
            call deleteAttribute(map,"HISTORY")
          else
            call message(AppMsg,Noisy,' ... No HISTORY keyword.')
          endif
          if (hasAttribute(map,"COMMENT")) then
            call deleteAttribute(map,"COMMENT")
          else
            call message(AppMsg,Noisy,' ... No COMMENT keyword.')
          endif
          call release(map)
          call release(outmask)
          call release(badpix)
        endif ! hasBADPIX
      enddo   ! loop over iccd=0,nccd
    endif     ! badpix_from_events
    if (.not. offsets_from_events) then
      call warning('noOFFSET','Cannot reliably handle OFFSETS, not included in FLAG map.')
    endif
  endif       ! outputmask requested

!
  if (is_epn) call CAL_setState(infile)
!
  call release(events)
  call release(infile)

  if (allocated(map_array)) deallocate(map_array)
!
!-----------------------------------------------------------------------
!
!     accumulate the distributions
!
  call message(AppMsg,Sparse,' ')
  call message(AppMsg,Verbose,' Pattern statistics for recognized events:')

  ks = 0
  kd = 0
  kt = 0
  kq = 0
  ymax = 0.0
  ymin = 1.0e+7
!
  do ibin = 1, maxbin
    ks = ks + khis(1,ibin)
    kd = kd + khis(2,ibin)
    kt = kt + khis(3,ibin)
    kq = kq + khis(4,ibin)
    khis(maxpat_plus_2,ibin) = &
                khis(1,ibin) + khis(2,ibin) + khis(3,ibin) + khis(4,ibin)
  enddo
!
  k = ks + kd + kt + kq
!
  hs = real(ks) / real(k)
  hd = real(kd) / real(k)
  ht = real(kt) / real(k)
  hq = real(kq) / real(k)
!
  write(errstr,'(4(7x,a1))') 's','d','t','q'
  call message(AppMsg,Sparse)
  write(errstr,'(4f8.4)') hs,hd,ht,hq
  call message(AppMsg,Sparse)

!
! compute the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... compute pattern distributions ...')
  do ipat = 1, maxpat_plus_1
    h = sigma * sigma
    s = 0.0
    i = 1
    j = 0
    k = 0
    do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
      i = i + 1
    enddo
    jm(1) = i
    if (i.gt.1) i = i - 1
    do while (i.lt.maxbin)
      i = i + 1
      s = s + real(khis(ipat,i))
      if (s*s.gt.h*s) then
        k = k + 1
        jcts(k) = s
        jp(k) = i
        s = 0.0
      endif
    enddo
    mbin = k

    if (mbin > 0) then
      do i = 2, mbin
        jm(i) = jp(i-1) + 1
      enddo

      do i = 1, mbin
        h = real(jp(i)-jm(i)) + 1.0
        sprb(i)  = real(jcts(i)) / h
      enddo
!
! store as histogram
!
      j = j+1
      xhis(ipat,j) = alog10((real(jm(1))-0.5)*conv)
      y = alog10(sprb(1))
      yhis(ipat,j) = y
      j = j+1
      xhis(ipat,j) = alog10((real(jp(1))+0.5)*conv)
      yhis(ipat,j) = y
      ymax = max(ymax,y)
      ymin = min(ymin,y)
      do i = 2, mbin
        j = j+1
        xhis(ipat,j) = xhis(ipat,j-1) 
        y = alog10(sprb(i))
        yhis(ipat,j) = y
        j = j+1
        xhis(ipat,j) = alog10((real(jp(i))+0.5)*conv)
        yhis(ipat,j) = y
        ymax = max(ymax,y)
        ymin = min(ymin,y)
      enddo
    else
      write(errstr,'(a,i1)') &
          'not enough statistics for at least 1 bin for PAT = ',ipat
      call warning('sigmaTooLarge')
    endif
    nbin(ipat) = j
  enddo

!
! initialize the plot for pattern distributions:
!
  call message(AppMsg,Noisy,' ')
  call message(AppMsg,Noisy,' ... prepare plot input data set ...')

  if (index(trim(device),"PS") .gt. 0)  is_ps = .true.
  if (index(trim(device),"GIF") .gt. 0) is_gif = .true.

! statusPG = pgopen(trim(device))
  if (is_ps) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  elseif (is_gif) then
    call pgbegin(0,trim(outfilename)//trim(device),1,1)
  else
    call pgbegin(0,trim(device),1,1)
  endif
  call pgpaper(7.8,1.414)
!
! top bar with some text
!
  vxmin = 0.15 - 0.75*0.12
  vxmax = 0.90
  vymin = 0.98
  vymax = 1.00
  grey  = 0.95
  call pgscr(255,grey,grey,grey)
  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(0.0,1.0,0.0,1.0)
  call pgsci(255)
  call pgrect(0.0,1.0,0.0,1.0)
  call pgsci(1)
  call pgsch(0.5)
  call pgslw(1)
  call pgptext(0.99, 0.4, 0.0, 1.0, trim(outfilename))
  call pgptext(0.01, 0.4, 0.0, 0.0, & 
           'XMM '//trim(insttxt)// &
           ', pattern statistics: epatplot-'//trim(CodeVersion))
!
! username, date and time at bottom of plot
!
  call pgsch(1.0)
  call pgiden

  call message(AppMsg,Noisy,' ... prepare X and Y ranges ...')
  xmin = alog10(17.0*conv)
  xmax = alog10(4096.0*conv)
  xamin = alog10(17.0)
  xamax = alog10(4096.0)
!  ymin = max(alog10(1.e-3),real(int(alog10(ymin)-1)))
!  ymax = real(int(alog10(ymax)+1))
  ymin = int(max(-3.0,ymin))-1
  ymax = int(ymax)+1
  write(errstr,*) '... unreasonable ranges: XMIN,XMAX, YMIN,YMAX = ', &
     xmin,xmax,ymin,ymax
  if (ymin>ymax) then
    call message(AppMsg,Verbose)
    call message(AppMsg,Verbose,' ... YMIN has been set to YMAX-2 ...')
    ymin = ymax - 2
  else
    call message(AppMsg,Noisy+1)  
  endif

  xsiz = xmax - xmin
  xasiz = xamax - xamin
  ysiz = ymax - ymin
  vxmin = 0.15
  vxmax = 0.90
  vymin = 0.52
  vymax = 0.90
  call message(AppMsg,Noisy+1,' ... call pgvport(vxmin,vxmax,vymin,vymax) ...')
  call pgvport(vxmin,vxmax,vymin,vymax)
  call message(AppMsg,Noisy+1,' ... call pgwindow(xmin,xmax,ymin,ymax) ...')
  call pgwindow(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(255) ...')
  call pgsci(255)
  call message(AppMsg,Noisy+1,' ... call pgrect(xmin,xmax,ymin,ymax) ...')
  call pgrect(xmin,xmax,ymin,ymax)
  call message(AppMsg,Noisy+1,' ... call pgsci(1) ...')
  call pgsci(1)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'LNV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('BCLST',0.0,0,'BCLST',0.0,0)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call message(AppMsg,Noisy+1,' ... call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel)) ...')
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
! call pgsch(1.25)
  call pgsch(1.12)
  call pgslw(4)
  xtxt = xmin - xsiz*0.12
  ytxt = ymax + ysiz*0.12
  call message(AppMsg,Noisy+1,' ... call pgptext(INSTTXT) ...')
  call pgptext(xtxt, ytxt, 0.0, 0.0, 'XMM / '//trim(insttxt)//' '//trim(instmode)//' '//trim(filter))
  xtxt = xmax
  ytxt = ymax + ysiz*0.12
  call pgsch(0.65)
  call pgslw(3)
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(revlab))
! call pgsch(1.0)
  call pgslw(3)
  call pgsch(0.65)
  ytxt = ymax + ysiz*0.05
  xtxt = xmax
  call message(AppMsg,Noisy+1,' ... call pgptext(FILENAME) ...')
  call pgptext(xtxt, ytxt, 0.0, 1.0, trim(infilename))
  call pgsch(1.0)
  call pgslw(4)
  xtxt = xmin - xsiz*0.1
! ytxt = ymin + ysiz*0.100
  ytxt = ymin - ysiz*0.20*0.5
  do ipat = maxpat_plus_1,1,-1
    call pgsci(icol(ipat))
    ytxt = ytxt + ysiz*0.20
    call pgptext(xtxt, ytxt, 90.0, 0.5, trim(chpt(ipat))//'/adu')
  enddo
  xtxt = xmin + xsiz*0.05
  ytxt = ymin + ysiz*0.05
  call pgsci(icol(maxpat_plus_1))
  write(errstr,'(1x,a,f6.2,a)') 'Energy fraction lost in invalid patterns in [1.0-32.8] keV:',&
       e_sum(nband,maxpat_plus_1)/e_sum(nband,maxpat_plus_2)*100,'%'
  call pgslw(2)
  call pgsch(0.50)
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(errstr))
  call pgsci(1)
  call pgslw(1)
!
! plot the pattern distributions vs. channel:
!
  call message(AppMsg,Noisy,' ... plot pattern distributions ...')
  do ipat = 1, maxpat_plus_1
    j = 0
!
! plot as histogram
!
    if (nbin(ipat) .gt. 0) then
      call pgsci(icol(ipat))
      j = j+1
      xdat = xhis(ipat,j)
      ydat = yhis(ipat,j)
      call pgmove(xdat,ydat)
      j = j+1
      xdat = xhis(ipat,j)
      call pgdraw(xdat,ydat)
      do i = j+1, nbin(ipat), 2
        ydat = yhis(ipat,i)
        call pgdraw(xdat,ydat)
        xdat = xhis(ipat,i+1)
        call pgdraw(xdat,ydat)
      enddo
    endif
  enddo
!
! initialize the plot for pattern statistics:
!
  call message(AppMsg,Noisy,' ... plot pattern statistics ...')
  ymin = -0.05
  ymax = 1.05
  ysiz = ymax - ymin
  vymin = vymin - 0.45
  vymax = vymax - 0.45
  call pgvport(vxmin,vxmax,vymin,vymax)
  call pgwindow(xmin,xmax,ymin,ymax)
  call pgsci(255)
  call pgrect(xmin,xmax,ymin,ymax)
  call pgsci(1)
  call pgslw(3)
  call pgbox('BC',0.0,0,'BC',0.0,0)
  call pgbox('LN',0.0,0,' ',0.0,0)
  call pgsch(0.75)
  call pgbox(' ',0.0,0,'NV',0.0,0)
  call pgsch(1.0)
  call pgslw(1)
  call pgbox('A',0.0,0,' ',0.0,0)
  call pgbox('BCLST',0.0,0,'BCST',0.0,0)
! call pgslw(6)
  call pgslw(5)
  call pgsch(1.1)
  xtxt = (xmax + xmin)*0.5
  ytxt = ymin - ysiz*0.11
  call pgptext(xtxt, ytxt, 0.0, 0.5, trim(xlabel))
  call pgslw(3)
  call pgsch(0.7)
!  xtxt = xmin + xsiz*0.95
!  ytxt = ymin + ysiz*0.85
  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.47
  call pgptext(xtxt, ytxt, 0.0, 0.0, trim(RAWYTXT))
  call pgslw(5)
  call pgsch(1.1)
!  xtxt = xmin + xsiz*0.95
!  ytxt = ymin + ysiz*0.91
  xtxt = xmin + xsiz*0.025
  ytxt = ymin + ysiz*0.51
  if (ccdmode(1:2) .ne. "  ") then
!   if (calMode .ne. 5) then
      call pgptext(xtxt, ytxt, 0.0, 0.0, trim(ccdmode)//' Model')
!   else
!     call pgptext(xtxt, ytxt, 0.0, 1.0, trim(ccdmode)//' (no model yet)')
!   endif
  endif
  call pgslw(4)
  call pgsci(icol(4))
  xtxt = xmin - xsiz*0.1
  ytxt = ymin + ysiz*yt0
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'q')
  call pgsci(icol(3))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 't')
  call pgsci(icol(2))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 'd')
  call pgsci(icol(1))
  ytxt = ytxt + ysiz*yt1
  call pgptext(xtxt, ytxt, 90.0, 0.5, 's')
  if (is_EPN) then
    call pgsci(6)
    ytxt = ytxt + ysiz*yt1
    call pgptext(xtxt, ytxt, 90.0, 0.5, 's+d')
  endif
  call pgsci(1)
  call pgslw(1)
!
! determine binning for statistics plot
!
  call message(AppMsg,Noisy,' ... determine binning for statistics plot ...')
  ipat = maxpat_plus_2
  h = sigma * sigma * 10.0  * 10.0
  s = 0.0
  i = 1
  k = 0
  do while (khis(ipat,i).eq.0 .and. i.lt.maxbin)
    i = i + 1
  enddo
  jm(1) = i
  if (i.gt.1) i = i - 1
  do while (i.lt.maxbin)
    i = i + 1
    s = s + real(khis(ipat,i))
    if (s*s.gt.h*s) then
      k = k + 1
      jp(k) = i
      s = 0.0
    endif
  enddo
  mbin = k
  do i = 2, mbin
    jm(i) = jp(i-1) + 1
  enddo
!
! determine the fraction of patterns among the recognized photons:
!
  call message(AppMsg,Noisy,&
    ' ... determine the fraction of patterns among the recognized photons ...')
  do ipat = 1, maxpat
    do i = 1, mbin
      k1 = 0
      k2 = 0
      do ibin = jm(i), jp(i)
        k1 = k1 + khis(ipat,ibin)
        k2 = k2 + khis(maxpat_plus_2,ibin)
      enddo
      rhis(ipat,i) = real(k1) / amax1(real(k2),1.0)
    enddo
!
! plot as histogram
!
    call pgsci(icol(ipat))
    xdat = alog10((real(jm(1))-0.5)*conv)
    ydat = rhis(ipat,1)
    call pgmove(xdat,ydat)
    xdat = alog10((real(jp(1))+0.5)*conv)
    call pgdraw(xdat,ydat)
    do i = 2, mbin
      ydat = rhis(ipat,i)
      call pgdraw(xdat,ydat)
      xdat = alog10((real(jp(i))+0.5)*conv)
      call pgdraw(xdat,ydat)
    enddo
  enddo

  if (is_EPN) then
    call pgsci(6)
    xdat = alog10((real(jm(1))-0.5)*conv)
    ydat = rhis(1,1)+rhis(2,1)
    call pgmove(xdat,ydat)
    xdat = alog10((real(jp(1))+0.5)*conv)
    call pgdraw(xdat,ydat)
    do i = 2, mbin
      ydat = rhis(1,i)+rhis(2,i)
      call pgdraw(xdat,ydat)
      xdat = alog10((real(jp(i))+0.5)*conv)
      call pgdraw(xdat,ydat)
    enddo
  endif
!------------------------------------------------------------------------------
! plot the standard single and double pattern statistics 
! as used in the response matrix
!
  call message(AppMsg,Noisy,' ... model curve as used for DRM ...')
!
  pfds = CAL_getPatternFractionServer()
  if (is_mos) then
    CALstat_ccd = 1
    call CAL_setState(CALstat_inst,    &
       CALstat_ccd, CALstat_node, ccdModeId = CALstat_mode, &
       accuracyLevel = ACCURACY_HIGH)
  endif
!
  do ipat = 1,2
    model(ipat,1:maxbin) = 0.0
    if (is_epn) then
!     call patmod(model(ipat,1:maxbin), maxbin, calMode, dthresh, ipat, ypos)
      iy16 = nint(ypos)
      ix16 = nint(xpos)
      call CAL_patternFractionAt(pfds, ix16, iy16)
      pattFracChan => CAL_getPatternFractionInChannel(pfds, ipat-1)
      mmp = size(pattFracChan)
      write(errstr,*) 'Number of channels in pattern array: ',mmp
      call message(AppMsg,Verbose)
      model(ipat,1:maxbin) = pattFracChan(0:maxbin-1)
    else
!     call patmod_mos(model(ipat,1:maxbin), maxbin, energy, ipat)
      curve = CAL_getPatternFractionInEnergy(pfds, ipat-1)
      do i = 1,maxbin
        model(ipat,i) = CAL_curveValueAt(curve, dble(energy(i)))
      enddo
    endif
  enddo

!
! plot as line
!
  call message(AppMsg,Noisy,' ... plot as line ...')
  if (is_epn) then
    call pgwindow(xamin,xamax,ymin,ymax)
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 20
      xdat = alog10(real(i))
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 21, maxbin
        xdat = alog10(real(i))
        ydat = model(ipat,i)
        call pgdraw(xdat,ydat)
      enddo
    enddo
! s+d:
    call pgsci(6)
    i = 20
    xdat = alog10(real(i))
    ydat = model(1,i)+model(2,i)
    call pgmove(xdat,ydat)
    do i = 21, maxbin
      xdat = alog10(real(i))
      ydat = min(model(1,i)+model(2,i),1.0)
      call pgdraw(xdat,ydat)
    enddo
  else
    do ipat = 1,2
      call pgsci(icol(ipat))
      i = 2
!     call message(AppMsg,Noisy,' ... !EPN: starting point ...')
      xdat = alog10(energy(i))
      ydat = model(ipat,i)
      call pgmove(xdat,ydat)
      do i = 3, maxbin
        if (energy(i) > 0) then
          xdat = alog10(energy(i))
          ydat = model(ipat,i)
!         print*,i,energy(i),xdat,ydat
          call pgdraw(xdat,ydat)
        endif
      enddo
!     call message(AppMsg,Noisy,' ... !EPN: loop done...')
    enddo
  endif
  call message(AppMsg,Noisy,' ... call pgend ...')
  call pgend
!
! write out results into ASCII file which can be plotted with QDP
!
  if (enable_qdp) then
    call get_io_Unit(iw)
    open(iw,file=outqdpname,status='unknown',iostat=ierr)
    if (ierr .eq. 0) then
      write(iw,'(a,a)') '! ',trim(outqdpname)
      write(iw,'(a)')'read serr 1'
      write(iw,'(a)')'col 0 on 2 3'
      write(iw,'(a)')'label x ADU Channel'
      write(iw,'(a)')'label y Pattern fraction'
      write(iw,'(a)')'resc y -0.05 1.05'
      write(iw,'(a)')'log x'
      write(iw,'(a)') '! CH           CHmin  CHmax  s'//&
                                  '        d        t        q'
      write(iw,'(a)') '! -----        -----  -----  ------   '//&
                                    '------   ------   ------'
  !
      do i = 1, mbin
        h = (jp(i)+jm(i))*0.5
        he = (jp(i)-jm(i))*0.5
        write(iw,'(2f6.1,2i7,4f9.4)') h,he,jm(i),jp(i),&
            rhis(1,i),rhis(2,i),rhis(3,i),rhis(4,i)
      enddo
      close(iw)
      call message(AppMsg,Verbose,' ')
      write(errstr,*) 'QDP output written to file ' // trim(outqdpname)
    else
      write(errstr,*) 'QDP output file could not be opened:' // trim(outqdpname)
    endif
    call message(AppMsg,Verbose)
  else
    call message(AppMsg,Verbose,' ')
  endif

  write(errstr,*) 'PostScript output written to file ' // trim(outfilename)
  call message(AppMsg,Sparse)
  call message(AppMsg,Verbose,' ')

end subroutine epatplot


! =========================
! Program parameter reading
! =========================
subroutine getInParamValues(infilename, sigma, device, outdir, withflag, &
                            useplotfile, plotfile, xaxisadu, enable_qdp, &
                            enable_pum, userMode, enable_Y, userY, &
                            withoutputmask,outmaskname)

  integer            ,intent(out)   :: userMode
  real(kind=single)  ,intent(out)   :: sigma, userY
  logical(kind=bool) ,intent(out)   :: useplotfile,xaxisadu,enable_qdp,&
                                       withflag, enable_pum, enable_Y, &
                                       withoutputmask
  character(len=*)   ,intent(out)   :: infilename,device,outdir,plotfile,&
                                       outmaskname

  infilename  = trim(stringParameter('set'))
  sigma       = realParameter('sigma')
  userMode    = intParameter('usermode')
  enable_pum  = booleanParameter('withusermode')
  userY       = realParameter('userrawy')
  enable_Y    = booleanParameter('withuserrawy')
  withflag    = booleanParameter('withflag')
  withoutputmask = booleanParameter('withoutputmask')
  outmaskname = stringParameter("outmaskname")
  useplotfile = booleanParameter('useplotfile')
  device      = trim(stringParameter('device'))
  outdir      = trim(stringParameter('outdir'))
  plotfile    = trim(stringParameter('plotfile'))
  xaxisadu    = booleanParameter('xaxisadu')
  enable_qdp  = booleanParameter('withqdp')

  if (verbosity(AppMsg) > Noisy) then
    call message(AppMsg,Sparse,'List of parameters as read:')
    write(errstr,*) 'set             -> infilename        ',infilename
    call message(AppMsg,Sparse)
    write(errstr,*) 'sigma           -> sigma      ',sigma
    call message(AppMsg,Sparse)
    write(errstr,*) 'usermode        -> userMode     ',userMode
    call message(AppMsg,Sparse)
    write(errstr,*) 'withusermode    -> enable_pum   ',enable_pum
    call message(AppMsg,Sparse)
    write(errstr,*) 'userrawy        -> userY      ',userY
    call message(AppMsg,Sparse)
    write(errstr,*) 'withuserrawy    -> enable_Y     ',enable_Y
    call message(AppMsg,Sparse)
    write(errstr,*) 'withflag        -> withflag     ',withflag
    call message(AppMsg,Sparse)
    write(errstr,*) 'withoutputmask  -> withoutputmask   ',withoutputmask
    call message(AppMsg,Sparse)
    write(errstr,*) 'outmaskname     -> outmaskname   ',outmaskname
    call message(AppMsg,Sparse)
    write(errstr,*) 'useplotfile     -> useplotfile   ',useplotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'device          -> device     ',device
    call message(AppMsg,Sparse)
    write(errstr,*) 'outdir          -> outdir     ',outdir
    call message(AppMsg,Sparse)
    write(errstr,*) 'plotfile        -> plotfile    ',plotfile
    call message(AppMsg,Sparse)
    write(errstr,*) 'xaxisadu        -> xaxisadu      ',xaxisadu
    call message(AppMsg,Sparse)
    write(errstr,*) 'withqdp         -> enable_qdp        ',enable_qdp
    call message(AppMsg,Sparse)
  endif

! if (.NOT. xaxisadu) then
!   call message(AppMsg,Sparse,'Parameter setting XAXISADU=N not ' // &
!               'implemented yet, using XAXISADU=Y in the following.')
!   xaxisadu = .true.
! endif

end subroutine getInParamValues


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine getCanonicalName(infilename,canonicalname)

  implicit none
  character(len=*)   ,intent(in)    :: infilename
  character(len=*)   ,intent(out)   :: canonicalname
  integer                           :: ns,nd,nl,il

  nl = len(trim(infilename))
  do il = nl, 1, -1
    if (infilename(il:il) == '/') exit
  enddo
  ns = max(il,1)
! print*,'IL,NS = ',il,ns
  do il = nl, ns, -1
    if (infilename(il:il) == '.') exit
  enddo
  if (il <= ns) then
    nd = nl
  else
    nd = max(il,2)-1
  endif
! print*,'IL,ND = ',il,nD
  canonicalname = infilename(ns:nd) // '_pat'
! print*,'|',trim(canonicalname),'|'

end subroutine getCanonicalName


! ==========================================
! Create output filename from input filename
! ==========================================
subroutine stripSlash(infilename)

  implicit none
  character(len=*)   ,intent(inout) :: infilename
  integer                           :: nl,il,jl

  nl = len(trim(infilename))
  do il = nl, 1, -1
    if (infilename(il:il) == '/') then
      do jl = il-1, 1, -1
        if (infilename(jl:jl) == '/') then
          infilename(jl:nl) = infilename(jl+1:nl)//' '
        else
          exit
        endif
      enddo
    endif
  enddo
  nl = len(trim(infilename))
  do il = nl, 3, -1
    if (infilename(il-2:il) == '/./') then
      infilename(il-2:nl) = infilename(il:nl)//'  '
    endif
  enddo
  if (infilename(1:2) == './') then
    infilename(1:nl) = infilename(3:nl)//'  '
  endif
! print*,'|',trim(infilename),'|'

end subroutine stripSlash


! ==========================================================================
! suggested by Clive Page in SSC-LUX-TN-0005 (Issue 1.0, 1996 Sept 2)
! ==========================================================================
subroutine get_io_unit(iounit)

  implicit none
  integer, intent(out) :: iounit
  logical              :: lopen, lexist
  do iounit = 99,0,-1
    inquire(unit=iounit, exist=lexist, opened=lopen)
    if (lexist .and. .not. lopen) exit
  enddo
  
end subroutine get_io_unit


end module epatplot_mod
